"""This module contains the Prompt class, which represents a prompt."""

import uuid
from typing import Dict, List


class Prompt:
    """A class that represents a prompt.

    Attributes:
        name (str): The name of the prompt.
        prompt_id (str): The ID of the prompt.
        system (str): The system that generated the prompt.
        user (str): The user that the prompt is for.
        version (int): The version of the prompt.
        template_vars (List[str]): The template variables in the prompt.
        id (str): The unique identifier for the prompt.
        active (bool): Whether the prompt is active.
    """

    def __init__(
        self,
        name: str,
        system: str,
        user: str,
        template_vars: List[str],
        version: int = 1,
        id: str | None = None,
        active: bool = False,
    ):
        self.name = name
        self.id = id if id else str(uuid.uuid4())
        self.system = system
        self.user = user
        self.version = version
        self.template_vars = template_vars
        self.active = active

    def __repr__(self):
        return f"Prompt(id={self.id}, \
                name={self.name}, \
                version={self.version}, \
                active={self.active}, \
                system={self.system}, \
                user={self.user}), \
                template_vars={self.template_vars})"

    def to_dict(self):
        return {
            "name": self.name,
            "id": self.id,
            "active": self.active,
            "system": self.system,
            "user": self.user,
            "version": self.version,
            "template_vars": self.template_vars,
        }

    @classmethod
    def from_dict(cls, data: Dict) -> "Prompt":
        return cls(
            name=data["name"],
            id=data["id"],
            active=data["active"],
            system=data["system"],
            user=data["user"],
            template_vars=data["template_vars"],
            version=data["version"],
        )



"""This ui element represent the input, prompt and output of a callable step in the PromptMage."""

from nicegui import ui, run, app
from typing import List
from loguru import logger

from promptmage.mage import MageStep
from .styles import textbox_style


RUNNING_ICON = "run_circle"
NOT_RUNNING_ICON = "circle"
SUCCESS_RUN_ICON = "check_circle"


class InputOutputSection:

    def __init__(self, step: MageStep):
        self.step = step

        self.fields = {}

    @ui.refreshable
    def ui(self):
        ui.label("Inputs:").classes("font-bold")
        for param in self.step.signature.parameters.values():
            if param.name not in ["prompt", "model"]:
                with ui.row():
                    self.fields[param.name] = (
                        ui.textarea(
                            label=f"{param.name}",
                            value=self.step.input_values[param.name],
                        )
                        .classes(textbox_style)
                        .props("outlined")
                    )


def create_function_runner(step: MageStep):
    input_output_section = InputOutputSection(step)
    system_prompt_field = None
    user_prompt_field = None
    model_select = None
    result_field = None
    expansion_tab = ui.expansion(
        f"Step: {step.name}", group="steps", icon=f"{NOT_RUNNING_ICON}"
    ).classes("text-lg w-full border")
    # load prompt if available
    if step.prompt_name:
        prompt = step.get_prompt()
        # load all available prompts for the step
        prompts = [
            p for p in step.prompt_store.get_prompts() if p.name == step.prompt_name
        ]
    else:
        prompt = None
        prompts = []

    # run id given in app.storage, initialize with this data
    if app.storage.user.get("step_run_id"):
        step_run_id = app.storage.user.get("step_run_id")
        run_data = step.data_store.get_data(step_run_id)
        if run_data.step_name == step.name:
            prompt = run_data.prompt
            step.input_values = run_data.input_data
            step.result = run_data.output_data
            expansion_tab.props(f"icon={SUCCESS_RUN_ICON}")
            expansion_tab.update()
            del app.storage.user["step_run_id"]

    async def run_function():
        expansion_tab.props(f"icon={RUNNING_ICON}")
        expansion_tab.update()
        inputs = {
            name: field.value for name, field in input_output_section.fields.items()
        }
        if prompt is not None:
            prompt.system = system_prompt_field.value
            prompt.user = user_prompt_field.value
        if model_select:
            logger.info(f"Selected model: {model_select.value}")
            step.model = model_select.value
        _ = await run.io_bound(step.execute, **inputs, prompt=prompt)
        if step.one_to_many:
            num_results = len(step.result.results)
            expansion_tab.props(f"caption='{num_results} results'")
        expansion_tab.props(f"icon={SUCCESS_RUN_ICON}")
        expansion_tab.update()

    def set_prompt():
        nonlocal prompt
        system_prompt_field.update()
        user_prompt_field.update()
        prompt.system = system_prompt_field.value
        prompt.user = user_prompt_field.value
        step.set_prompt(prompt)

    def update_inputs():
        for name, field in input_output_section.fields.items():
            field.set_value(step.input_values[name])
            field.update()
        expansion_tab.props(f"icon={RUNNING_ICON}")
        expansion_tab.update()

    def update_results():
        newline = "\n\n"
        if isinstance(step.result, list):
            result_field.set_content(
                f"{[newline.join(result.results.values()) for result in step.result]}"
            )
        else:
            result_field.set_content(f"{newline.join(step.result.results.values())}")
        result_field.update()

        expansion_tab.props(f"icon={SUCCESS_RUN_ICON}")
        if isinstance(step.result, list):
            num_results = len(step.result)
            expansion_tab.props(f"caption='{num_results} results'")
        expansion_tab.update()

    def display_prompt(prompt_str: str):
        nonlocal prompt
        logger.info(f"Selected prompt: {prompt_str}")
        prompt_name, version = prompt_str.split(" - v")
        prompt = step.prompt_store.get_prompt(prompt_name, version=int(version))
        system_prompt_field.set_value(prompt.system)
        user_prompt_field.set_value(prompt.user)
        system_prompt_field.update()
        user_prompt_field.update()

    step.on_input_change(update_inputs)
    step.on_output_change(update_results)

    def build_ui():
        nonlocal user_prompt_field, system_prompt_field, result_field, expansion_tab, model_select
        with expansion_tab:
            ui.label(f"ID: {step.step_id}")
            with ui.column().classes("w-full"):
                with ui.row().classes("w-full"):
                    # show available models if available
                    if step.available_models:
                        model_select = ui.select(
                            step.available_models,
                            label="Select model",
                            value=step.model,
                        ).classes("w-1/3")
                    # show available prompts
                    prompt_select = ui.select(
                        [f"{p.name} - v{p.version}" for p in prompts],
                        label="Select prompt",
                        value=(
                            f"{prompt.name} - v{prompt.version}"
                            if prompt
                            else "No prompt selected"
                        ),
                        on_change=lambda event: display_prompt(event.value),
                    ).classes("w-1/3")

                with ui.row().classes("w-full"):
                    ui.label("Prompts:").classes("font-bold")
                    with ui.row().classes("grow"):
                        system_prompt_field = (
                            ui.textarea(
                                label="System prompt:",
                                value=(
                                    prompt.system if prompt else "No prompt supported"
                                ),
                            )
                            .classes(textbox_style)
                            .props("outlined")
                        )
                    with ui.row().classes("grow"):
                        user_prompt_field = (
                            ui.textarea(
                                label="User prompt:",
                                value=(
                                    prompt.user if prompt else "No prompt supported"
                                ),
                            )
                            .classes(textbox_style)
                            .props("outlined")
                        )
                with ui.row():
                    input_output_section.ui()

                with ui.row().classes("w-full justify-end"):
                    ui.button("Run", on_click=run_function)
                    ui.button("Save prompt", on_click=set_prompt)
                ui.separator()
                with ui.row().classes("w-full justify-between"):
                    ui.label("Result:").classes("font-bold")
                    ui.button(
                        "Copy to clipboard",
                        on_click=lambda: ui.clipboard.write(
                            step.result or "No result available"
                        ),
                    )
                result_field = (
                    ui.markdown(f"{step.result}" if step.result else "")
                    .style("margin-top: 20px;")
                    .classes("color-black dark:color-white")
                )

    return build_ui



import json
from typing import List
from loguru import logger
from slugify import slugify
from nicegui import ui, app

from promptmage import PromptMage, RunData
from .styles import label_with_icon


def create_runs_view(mage: PromptMage):

    datasets = mage.data_store.backend.get_datasets()

    side_panel = (
        ui.element("div")
        .style(
            "position: fixed; top: 0; right: 0; width: 50%; height: 100%; transform: translateX(100%); transition: transform 0.3s ease; z-index: 1000; overflow-y: auto;"
        )
        .classes("bg-gray-100 dark:bg-slate-800")
    )
    # compare runs dialog
    compare_dialog = ui.dialog().props("full-width")

    def use_run_in_playground(step_run_id):
        app.storage.user["step_run_id"] = step_run_id
        ui.navigate.to(f"/{slugify(mage.name)}")

    # Function to show the side panel with detailed information
    def show_side_panel(run_data: RunData):
        side_panel.clear()
        with side_panel:
            ui.button(">>", on_click=hide_side_panel).style(
                "margin: 20px; margin-bottom: 0px; margin-top: 100px;"
            )
            with ui.card().style(
                "padding: 20px; margin-right: 20px; margin-top: 20px; margin-bottom: 20px; margin-left: 20px"
            ):
                # display run data
                with ui.row().classes("w-full"):
                    with ui.column().classes("gap-0"):
                        ui.label("Step name").classes("text-sm text-gray-500")
                        ui.label(f"{run_data.step_name}").classes("text-2xl")
                    ui.space()
                    with ui.column().classes("gap-0 items-center"):
                        ui.label("Status").classes("text-sm text-gray-500")
                        ui.chip(
                            f"{run_data.status}",
                            icon="",
                            color=f"{'green' if run_data.status == 'success' else 'red'}",
                        ).props("outline square")
                with ui.row().classes("w-full"):
                    with ui.column().classes("gap-0"):
                        label_with_icon(
                            "Execution time:", icon="hourglass_bottom"
                        ).classes("text-sm text-gray-500")
                        label_with_icon("Run At:", icon="o_schedule").classes(
                            "text-sm text-gray-500"
                        )
                        label_with_icon("Model:", icon="o_psychology").classes(
                            "text-sm text-gray-500"
                        )
                        label_with_icon("Step Run ID:", icon="o_info").classes(
                            "text-sm text-gray-500"
                        )
                        label_with_icon("Run ID:", icon="o_info").classes(
                            "text-sm text-gray-500"
                        )
                    with ui.column().classes("gap-0"):
                        ui.label(
                            f"{run_data.execution_time if run_data.execution_time else 0.0:.2f}s"
                        )
                        ui.label(f"{run_data.run_time[:19]}")
                        ui.label(f"{run_data.model}")
                        ui.label(f"{run_data.step_run_id}")
                        ui.label(f"{run_data.run_id}")

                if run_data.prompt:
                    ui.label("Prompt:").classes("text-lg")
                    with ui.row().classes("w-full"):
                        with ui.column().classes("gap-0"):
                            ui.label("Version:").classes("text-sm text-gray-500")
                            ui.label("ID:").classes("text-sm text-gray-500")
                            ui.label("System Prompt:").classes("text-sm text-gray-500")
                            ui.label("User Prompt:").classes("text-sm text-gray-500")
                        with ui.column().classes("gap-0 w-1/2"):
                            ui.label(f"{run_data.prompt.version}")
                            ui.label(f"{run_data.prompt.id}")
                            ui.label(f"{run_data.prompt.system}")
                            ui.label(f"{run_data.prompt.user}")

                ui.label("Input Data:").classes("text-lg")
                for key, value in run_data.input_data.items():
                    ui.markdown(f"**{key}**")
                    ui.markdown(f"{value}")
                ui.label("Output Data:").classes("text-lg")
                try:
                    for key, value in run_data.output_data.items():
                        ui.markdown(f"**{key}**")
                        ui.markdown(f"{value}")
                except AttributeError:
                    ui.markdown(f"{run_data.output_data}")
                with ui.row():
                    ui.button(
                        "Use in playground",
                        on_click=lambda: use_run_in_playground(run_data.step_run_id),
                    )
        side_panel.style("transform:translateX(0%);")
        side_panel.update()

    # Function to hide the side panel
    def hide_side_panel():
        side_panel.clear()
        side_panel.style("transform:translateX(100%);")
        side_panel.update()

    # rating function
    def rate_run(step_run_id, rating):
        logger.info(f"Rating run {step_run_id} with {rating}")
        ui.notify("Not implemented yet. Run rated successfully.")

    def build_ui():
        runs: List[RunData] = mage.get_run_data()

        def display_comparison():
            selected_runs = table.selected
            if len(selected_runs) < 2:
                ui.notify("Please select at least two runs to compare.")
                return
            if len(selected_runs) > 5:
                ui.notify("Please select at most five runs to compare.")
                return
            status_success = all([r["status"] == "success" for r in selected_runs])
            if not status_success:
                ui.notify("Please select only successful runs to compare.")
                return
            # check if all selected runs are from the same step
            step_names = set([r["name"] for r in selected_runs])
            if len(step_names) > 1:
                ui.notify("Please select runs from the same step to compare.")
                return
            compare_dialog.clear()
            with compare_dialog, ui.card():
                ui.label(f"Compare Runs for step {step_names.pop()}").classes(
                    "text-2xl"
                )
                with ui.row().style(
                    "display: flex; width: 100%; align-items: stretch;"
                ):
                    for run_data in selected_runs:
                        # get the results for the selected run
                        run: RunData = [
                            run
                            for run in runs
                            if run.step_run_id == run_data["step_run_id"]
                        ][0]
                        with ui.column().style("flex: 1;"):
                            with ui.card().style(
                                "flex-grow: 1; display: flex; flex-direction: column;"
                            ):
                                ui.label(f"step_run_id: {run_data['step_run_id']}")
                                ui.markdown(f"{run.output_data}")
                                with ui.button_group().style("margin-top: auto;"):
                                    ui.button(
                                        icon="thumb_up",
                                        on_click=lambda: rate_run(
                                            run.step_run_id, "up"
                                        ),
                                    )
                                    ui.button(
                                        icon="thumb_down",
                                        on_click=lambda: rate_run(
                                            run.step_run_id, "down"
                                        ),
                                    )

                ui.button("Close", on_click=compare_dialog.close)
            compare_dialog.open()

        def add_to_dataset():
            if dataset_select.value is None:
                ui.notify("Please select a dataset to add the runs to.")
                return
            dataset = datasets[dataset_select.value]
            selected_runs = table.selected
            for run in selected_runs:
                mage.data_store.backend.add_datapoint_to_dataset(
                    run["step_run_id"], dataset.id
                )
                logger.info(f"Added run {run['step_run_id']} to dataset {dataset.id}")
            ui.notify(
                f"Added {len(selected_runs)} runs to dataset {dataset.name} successfully."
            )

        # Main UI setup
        with ui.card().style("padding: 20px").classes("w-full"):
            with ui.row().classes("w-full"):
                ui.button("Compare Runs", on_click=display_comparison).style(
                    "margin-bottom: 20px"
                )
                ui.button("Add to dataset", on_click=add_to_dataset).style(
                    "margin-bottom: 20px"
                )
                dataset_select = ui.select(
                    {idx: f"{d.name}-{d.id}" for idx, d in enumerate(datasets)},
                    value=None,
                    label="Select Dataset",
                ).classes("w-1/3")
            # Create a table with clickable rows
            columns = [
                {
                    "name": "run_time",
                    "label": "Run At",
                    "field": "run_time",
                    "sortable": True,
                },
                {
                    "name": "status",
                    "label": "Status",
                    "field": "status",
                    "sortable": True,
                },
                {"name": "name", "label": "Name", "field": "name", "sortable": True},
                {
                    "name": "execution_time",
                    "label": "Execution Time (in seconds)",
                    "field": "execution_time",
                    "sortable": True,
                },
                {
                    "name": "run_id",
                    "label": "run_id",
                    "field": "run_id",
                },
                {"name": "step_run_id", "label": "step_run_id", "field": "step_run_id"},
            ]

            rows = [
                {
                    "run_id": run_data.run_id,
                    "step_run_id": run_data.step_run_id,
                    "name": run_data.step_name,
                    "status": run_data.status,
                    "run_time": run_data.run_time[:19],
                    "execution_time": (
                        round(run_data.execution_time, 3)
                        if run_data.execution_time
                        else -1
                    ),
                }
                for run_data in runs
            ]

            table = ui.table(
                columns=columns,
                rows=rows,
                selection="multiple",
                row_key="step_run_id",
                pagination={
                    "rowsPerPage": 20,
                    "sortBy": "run_time",
                    "page": 1,
                    "descending": True,
                },
            )
            table.add_slot(
                "body-cell-status",
                """
                <q-td key="status" :props="props">
                    <q-badge :color="props.value === 'failed' ? 'red' : 'green'">
                        {{ props.value }}
                    </q-badge>
                </q-td>
                """,
            )

            def on_row_click(event):
                selected_run_index = event.args[-2]["step_run_id"]
                show_side_panel(
                    run_data=[r for r in runs if r.step_run_id == selected_run_index][
                        -1
                    ]
                )

            table.on("rowClick", on_row_click)

    return build_ui



from nicegui import ui
from typing import List
from slugify import slugify
from loguru import logger

from promptmage import PromptMage


def dataset_card(dataset, flow):
    with ui.card().style("padding: 20px; margin: 10px;"):
        ui.label(f"{dataset.name}").classes("text-lg")
        ui.button(
            "Go to dataset",
            on_click=lambda: ui.navigate.to(
                f"/evaluation/{slugify(flow.name)}/{dataset.id}"
            ),
        )
        ui.separator()
        datapoints = flow.data_store.backend.get_datapoints(dataset.id)
        ui.chip(f"{len(datapoints)} Datapoints", icon="check_circle").props("square")
        ui.separator()
        ui.button(
            "Delete",
            on_click=lambda: delete_dataset(dataset.id, flow),
        ).style("color: red;")


def create_dataset(name: str, flow: PromptMage):
    flow.data_store.backend.create_dataset(name)
    logger.info(f"Created dataset: {name}")


def delete_dataset(dataset_id: str, flow: PromptMage):
    flow.data_store.backend.delete_dataset(dataset_id)
    logger.info(f"Deleted dataset: {dataset_id}")


def create_grid(elements, flow, columns: int = 4):
    elements.insert(0, None)
    rows = len(elements) // columns + (1 if len(elements) % columns > 0 else 0)
    if rows == 0:
        rows = 1
    for i in range(rows):
        with ui.row().classes("justify-center"):
            for j in range(columns):
                if i == j == 0:
                    with ui.card().style("padding: 20px; margin: 10px;"):
                        ui.button(
                            "Create dataset",
                            on_click=lambda: create_dataset("New Dataset", flow),
                        )

                else:
                    index = i * columns + j
                    if index < len(elements):
                        dataset_card(elements[index], flow=flow)


def build_evaluation_page(flow: PromptMage):
    available_datasets = flow.data_store.backend.get_datasets()
    with ui.column().classes("items-center"):
        create_grid(available_datasets, flow=flow, columns=5)



import uuid
import time
import inspect
from typing import Callable, List
from loguru import logger

from .prompt import Prompt
from .run_data import RunData
from .result import MageResult
from .storage import PromptStore, DataStore


class MageStep:
    """A class to represent a step in a PromptMage instance.

    Attributes:
        step_id (str): The unique identifier for the step.
        name (str): The name of the step.
        func (Callable): The function to execute for the step.
        signature (inspect.Signature): The signature of the function.
        prompt_store (PromptStore): The prompt store to use for storing prompts.
        data_store (DataStore): The data store to use for storing data.
        prompt_name (str): The name of the prompt associated with the step.
        depends_on (str): The name of the step that this step depends on.
        initial (bool): Whether the step is an initial step.
        one_to_many (bool): Whether the step is a one-to-many step.
        many_to_one (bool): Whether the step is a many-to-one step.
        model (str): The model to use for the step.
        available_models (List[str]): The available models for the step.
        pass_through_inputs (List[str]): The inputs to pass through to the next step.
    """

    def __init__(
        self,
        name: str,
        func: Callable,
        prompt_store: PromptStore,
        data_store: DataStore,
        prompt_name: str | None = None,
        depends_on: str | None = None,
        initial: bool = False,
        one_to_many: bool = False,
        many_to_one: bool = False,
        model: str | None = None,
        available_models: List[str] | None = None,
        pass_through_inputs: List[str] | None = None,
    ):
        self.step_id = str(uuid.uuid4())
        self.name = name
        self.func = func
        self.signature = inspect.signature(func)
        self.prompt_store = prompt_store
        self.data_store = data_store
        self.prompt_name = prompt_name
        self.depends_on = depends_on
        self.initial = initial
        self.one_to_many = one_to_many
        self.many_to_one = many_to_one
        self.model = model
        self.available_models = available_models
        self.pass_through_inputs = pass_through_inputs

        # store inputs and results
        self.input_values = {}
        self.result = None

        # Initialize input values with default parameter values
        for param in self.signature.parameters.values():
            if param.name in ["prompt", "model"]:
                continue
            if param.default is not inspect.Parameter.empty:
                self.input_values[param.name] = param.default
            else:
                self.input_values[param.name] = None

        # callbacks for the frontend
        self._input_callbacks = []
        self._output_callbacks = []

    def execute(
        self, prompt: Prompt | None = None, active: bool | None = None, **inputs
    ):
        """Execute the step with the given inputs."""
        logger.info(f"Executing step: {self.name}...")
        multi_input_param = None
        # set the inputs
        logger.info(f"Setting inputs: {inputs.keys()}")
        for key, value in inputs.items():
            if isinstance(value, list):
                multi_input_param = key
            self.input_values[key] = value
        # set the model
        if self.model:
            self.input_values["model"] = self.model
        # get the prompt and set it if exists
        if self.prompt_name:
            if prompt:
                self.input_values["prompt"] = prompt
            else:
                prompt = self.get_prompt(active=active)
                self.input_values["prompt"] = prompt
        else:
            prompt = None
        # run the input callbacks
        for callback in self._input_callbacks:
            callback()
        # execute the function and store the result
        start_time = time.time()
        try:
            if self.one_to_many:
                logger.info("Executing step one-to-many")
                self.result = []
                if multi_input_param:
                    param_values = self.input_values[multi_input_param]
                    for value in param_values:
                        self.input_values[multi_input_param] = value
                        self.result.append(self.func(**self.input_values))
                    self.input_values[multi_input_param] = param_values
                else:
                    raise ValueError(
                        "One-to-many step requires a list input parameter."
                    )
            elif self.many_to_one:
                logger.info("Executing step many-to-one")
                self.result = self.func(**self.input_values)
            else:
                logger.info("Executing step normally")
                self.result = self.func(**self.input_values)
            status = "success"
        except Exception as e:
            logger.error(f"Error executing step: {e}")
            self.result = MageResult(error=f"Error: {e}")
            status = "failed"
        execution_time = time.time() - start_time
        # store the run data
        self.store_run(prompt=prompt, status=status, execution_time=execution_time)
        # run the output callbacks
        for callback in self._output_callbacks:
            callback()
        logger.info(f"Step {self.name} executed successfully.")
        return self.result

    def __repr__(self):
        return (
            f"Step(step_id={self.step_id}, "
            f"name={self.name}, "
            f"prompt_name={self.prompt_name}, "
            f"depends_on={self.depends_on})"
        )

    def get_prompt(
        self, version: int | None = None, active: bool | None = None
    ) -> Prompt:
        return self.prompt_store.get_prompt(self.prompt_name, version, active)

    def set_prompt(self, prompt: Prompt):
        prompt.id = str(uuid.uuid4())
        prompt.version = prompt.version + 1
        self.prompt_store.store_prompt(prompt)

    def store_run(
        self,
        prompt: Prompt | None = None,
        status: str = "success",
        execution_time: float = 0.0,
    ):
        """Store the run data in the data store."""
        if self.data_store:
            run_data = RunData(
                step_name=self.name,
                prompt=prompt if self.prompt_name else None,
                input_data={
                    k: v
                    for k, v in self.input_values.items()
                    if k not in ["prompt", "model"]
                },
                output_data=(
                    [r.results for r in self.result]
                    if isinstance(self.result, list)
                    else self.result.results
                ),
                status=status,
                model=self.model,
                execution_time=execution_time,
            )
            self.data_store.store_data(run_data)

    def on_input_change(self, callback):
        self._input_callbacks.append(callback)

    def on_output_change(self, callback):
        self._output_callbacks.append(callback)



"""Frontend module for PromptMage."""

from fastapi import FastAPI
from nicegui import ui
from typing import List
from slugify import slugify

from promptmage import PromptMage

from .components import theme
from .components.runs_page import create_runs_view
from .components.prompts_page import create_prompts_view
from .components.overview_page import build_overview_page
from .components.flow_page import build_flow_page
from .components.evaluation_page import build_evaluation_page
from .components.dataset_page import build_dataset_page


class PromptMageFrontend:
    """A class that creates a frontend for a PromptMage instance."""

    def __init__(self, flows: List[PromptMage]):
        self.flows = flows
        self.flows_dict = {slugify(flow.name): flow for flow in flows}
        self.current_flow = self.flows[0]

    def init_from_api(self, fastapi_app: FastAPI) -> None:
        """Initialize the frontend from a FastAPI application."""

        @ui.page("/", title="PromptMage")
        def main_page():
            with theme.frame(
                "Welcome to the PromptMage", flow_name=slugify(self.current_flow.name)
            ):
                build_overview_page(self.flows)

        @ui.page("/{flow_name}", title="PromptMage - Flow")
        def flow_page(flow_name: str):
            self.current_flow = self.flows_dict[flow_name]
            with theme.frame("Playground", flow_name=flow_name):
                build_flow_page(self.flows_dict[flow_name])

        @ui.page("/runs/{flow_name}", title="PromptMage - Runs")
        def runs_page(flow_name: str):
            self.current_flow = self.flows_dict[flow_name]
            with theme.frame(f"Runs Overview - {flow_name}", flow_name=flow_name):
                create_runs_view(self.current_flow)()

        @ui.page("/prompts/{flow_name}", title="PromptMage - Prompts")
        def prompts_page(flow_name: str):
            self.current_flow = self.flows_dict[flow_name]
            with theme.frame(f"Prompts Overview - {flow_name}", flow_name=flow_name):
                create_prompts_view(self.current_flow)()

        @ui.page("/evaluation/{flow_name}", title="PromptMage - Evaluation")
        def evaluation_page(flow_name: str):
            self.current_flow = self.flows_dict[flow_name]
            with theme.frame(f"Evaluation - {flow_name}", flow_name=flow_name):
                build_evaluation_page(self.current_flow)

        @ui.page(
            "/evaluation/{flow_name}/{dataset_id}",
            title="PromptMage - Evaluation Dataset",
        )
        def evaluation_dataset_page(flow_name: str, dataset_id: str):
            self.current_flow = self.flows_dict[flow_name]
            with theme.frame(
                f"Evaluation - {flow_name} - {dataset_id}", flow_name=flow_name
            ):
                build_dataset_page(self.current_flow, dataset_id)()

        ui.run_with(
            fastapi_app,
            mount_path="/gui",  # NOTE this can be omitted if you want the paths passed to @ui.page to be at the root
            storage_secret="pick your private secret here",  # NOTE setting a secret is optional but allows for persistent storage per user
            dark=True,
            favicon="🧙",
        )



from contextlib import contextmanager
from loguru import logger
from nicegui import ui, app


from .menu import menu


@contextmanager
def frame(navigation_title: str, flow_name: str = None):
    """Custom page frame to share the same styling and behavior across all pages"""
    dark_mode_icon = None
    # load dark mode from local storage
    dark_value = app.storage.user.get("dark_mode", True)
    dark = ui.dark_mode(value=dark_value)

    def toggle_dark_mode():
        """Toggle dark mode on and off."""
        if dark.value:
            dark.disable()
            dark_mode_icon.name = "dark_mode"
        else:
            dark.enable()
            dark_mode_icon.name = "light_mode"
        # store dark mode in local storage
        app.storage.user["dark_mode"] = dark.value
        dark_mode_icon.update()

    # ui.colors(
    #    primary="#6E93D6", secondary="#53B689", accent="#111B1E", positive="#53B689"
    # )
    ui.colors(
        primary="#166088",
        secondary="#8fb9d2",
        accent="#FFFD98",
        dark="#13140b",
        positive="#21ba45",
        negative="#c10015",
        info="#bde4a7",
        warning="#f2c037",
    )

    with ui.header(elevated=True):
        with ui.button(on_click=lambda event: ui.navigate.to("/")).props(
            "flat color=white"
        ).classes("p-0"):
            ui.avatar("img:/static/promptmage-logo.png", square=True)
        ui.label(navigation_title).classes("text-2xl self-center")
        ui.space()
        menu(flow_name=flow_name)
        with ui.button(
            on_click=toggle_dark_mode,
        ).props(
            "flat small color=white"
        ).classes("self-center"):
            dark_mode_icon = ui.icon("light_mode" if dark.value else "dark_mode")

    with ui.column().classes("self-center w-full gap-0"):
        yield



"""This module contains the InMemoryBackend class, which implements a simple in-memory storage backend for prompts."""

from typing import Dict

from promptmage.prompt import Prompt
from promptmage.run_data import RunData
from promptmage.storage import StorageBackend
from promptmage.exceptions import PromptNotFoundException


class InMemoryPromptBackend(StorageBackend):
    """A simple in-memory storage backend for prompts."""

    def __init__(self):
        self.prompts: Dict[str, Prompt] = {}

    def store_prompt(self, prompt: Prompt):
        """Store a prompt in memory."""
        self.prompts[prompt.name] = prompt.to_dict()

    def get_prompt(self, prompt_name: str) -> str:
        """Retrieve a prompt from memory."""
        if prompt_name not in self.prompts:
            raise PromptNotFoundException(f"Prompt with name {prompt_name} not found.")
        return Prompt.from_dict(self.prompts.get(prompt_name))

    def get_prompts(self) -> Dict:
        """Retrieve all prompts from memory."""
        return self.prompts


class InMemoryDataBackend(StorageBackend):
    """A simple in-memory storage backend for data."""

    def __init__(self):
        self.data: Dict[str, RunData] = {}

    def store_data(self, run: RunData):
        """Store data in memory."""
        self.data[run.run_id] = run.to_dict()

    def get_data(self, run_id: str) -> str:
        """Retrieve data from memory."""
        return RunData.from_dict(self.data.get(run_id))

    def get_all_data(self) -> Dict:
        """Retrieve all data from memory."""
        return self.data



from nicegui import ui


def menu(flow_name: str) -> None:
    with ui.button_group().classes("self-center"):
        ui.button("Overview", on_click=lambda: ui.navigate.to("/")).classes(
            replace="text-white"
        )
        ui.button(
            "Playground", on_click=lambda: ui.navigate.to(f"/{flow_name}")
        ).classes(replace="text-white")
        ui.button(
            "Runs", on_click=lambda: ui.navigate.to(f"/runs/{flow_name}")
        ).classes(replace="text-white")
        ui.button(
            "Prompts", on_click=lambda: ui.navigate.to(f"/prompts/{flow_name}")
        ).classes(replace="text-white")
        ui.button(
            "Evaluation", on_click=lambda: ui.navigate.to(f"/evaluation/{flow_name}")
        ).classes(replace="text-white")



from nicegui import ui
from typing import List
from slugify import slugify

from promptmage import PromptMage


def flow_card(flow: PromptMage):
    with ui.card().style("padding: 20px; margin: 10px;"):
        ui.button(
            f"{flow.name}", on_click=lambda: ui.navigate.to(f"/{slugify(flow.name)}")
        )
        ui.separator()
        ui.chip(f"{len(flow.steps)} Steps", icon="run_circle").props("square")
        runs = flow.get_run_data()
        ui.chip(f"{len(runs)} Runs", icon="check_circle").props("square")


def create_grid(elements, columns=4):
    rows = len(elements) // columns + (1 if len(elements) % columns > 0 else 0)
    for i in range(rows):
        with ui.row():
            for j in range(columns):
                index = i * columns + j
                if index < len(elements):
                    flow_card(elements[index])


def build_overview_page(flows: List[PromptMage]):
    create_grid(flows, columns=4)



from nicegui import ui
from loguru import logger

from promptmage import PromptMage
from promptmage.step import MageStep
from .main_runner import create_main_runner
from .step_runner import create_function_runner


def render_mermaid_diagram(execution_list: list) -> str:
    """
    Generate a Mermaid diagram from the self.execution_results.

    Returns:
        str: The Mermaid diagram code.
    """
    diagram = ["graph TD"]
    graph_events = ""
    all_nodes = set()  # To keep track of all nodes
    nodes_with_outgoing_edges = set()  # To track nodes with outgoing edges
    id_to_step_name = {}  # Mapping of result_id to step name
    id_to_result = {}  # Mapping of result_id to result

    for result in execution_list:
        previous_ids = result["previous_result_ids"]
        current_id = result["current_result_id"]
        step_name = result["step"]

        # Store the mapping from id to step name
        id_to_step_name[current_id] = step_name

        # Store the mapping from id to result
        id_to_result[current_id] = result["results"]

        all_nodes.add(current_id)

        if previous_ids:
            for prev_id in previous_ids:
                diagram.append(f"    {prev_id} --> {current_id}")
                graph_events += (
                    f'click {prev_id} call emitEvent("graph_click", {prev_id})\n'
                )
                all_nodes.add(prev_id)
                nodes_with_outgoing_edges.add(prev_id)
        else:
            diagram.append(f"    start --> {current_id}")
            nodes_with_outgoing_edges.add("start")

    # Identify terminal nodes (nodes that do not have outgoing edges)
    terminal_nodes = all_nodes - nodes_with_outgoing_edges

    # Add 'END' for terminal nodes
    for node in terminal_nodes:
        graph_events += f'click {node} call emitEvent("graph_click", {node})\n'
        diagram.append(f"    {node} --> END")

    # Add labels for nodes using their step names
    for node_id, step_name in id_to_step_name.items():
        diagram.append(f"    {node_id}({step_name})")

    return "\n".join(diagram) + "\n" + graph_events, id_to_step_name, id_to_result


@ui.refreshable
def execution_graph(flow: PromptMage):
    with ui.dialog() as dialog, ui.card():
        ui.label("Execution Result will be shown here.")

    if flow.execution_results:
        graph, id_to_step_name, id_to_result = render_mermaid_diagram(
            flow.execution_results
        )

        def node_dialog(id: str):
            dialog.clear()
            with dialog, ui.card().classes("w-512 h-128"):
                with ui.row().classes("w-full justify-between"):
                    ui.label(f"Result for step '{id_to_step_name[id]}'").classes(
                        "text-lg"
                    )
                    ui.space()
                    ui.button("Close", on_click=dialog.close)
                ui.markdown(
                    "\n\n".join(
                        f"{variable}:\n{result}"
                        for variable, result in id_to_result[id].items()
                    )
                )
            dialog.open()

        ui.mermaid(graph, config={"securityLevel": "loose"}).classes("w-2/3")
        ui.on("graph_click", lambda e: node_dialog(e.args))
    elif flow.is_running:
        ui.spinner("puff", size="xl")
    else:
        ui.label("No execution results available.")


def build_flow_page(flow: PromptMage):
    with ui.row().classes("w-full gap-0"):
        with ui.splitter().classes("w-full p-0") as splitter:
            with splitter.before:
                with ui.column().classes("w-full items-center"):
                    # Create a card for the mage
                    with ui.card().classes("w-full"):
                        ui.label(f"Flow: {flow.name}").classes("text-xl")
                        with create_main_runner(flow, execution_graph):
                            # Create a card for each step
                            with ui.column().classes("w-full"):
                                step: MageStep
                                for step in flow.steps.values():
                                    create_function_runner(step)()
            with splitter.after:
                # Create a card for the execution graph
                with ui.column().classes("w-full items-center pl-4"):
                    with ui.card().classes("w-full items-center"):
                        ui.label("Execution graph").classes("text-xl")
                        execution_graph(flow)



class PromptNotFoundException(Exception):
    """Raised when a prompt is not found in the backend."""

    def __init__(self, prompt_id: str):
        self.prompt_id = prompt_id
        super().__init__(f"Prompt with ID {prompt_id} not found.")


class DataNotFoundException(Exception):
    """Raised when data is not found in the backend."""

    def __init__(self, data_id: str):
        self.data_id = data_id
        super().__init__(f"Data with ID {data_id} not found.")



from nicegui import ui

from loguru import logger
from typing import List

from promptmage import PromptMage
from promptmage.run_data import RunData


def build_dataset_page(flow: PromptMage, dataset_id: str):
    dataset = flow.data_store.backend.get_dataset(dataset_id)
    ui.label(f"Dataset name: {dataset.name} - id: {dataset_id}").classes("text-2xl")

    datapoints = flow.data_store.backend.get_datapoints(dataset_id)

    side_panel = ui.element("div").style(
        "position: fixed; top: 0; right: 0; width: 50%; height: 100%; background-color: #f0f0f0; transform: translateX(100%); transition: transform 0.3s ease; z-index: 1000; overflow-y: auto;"
    )
    # compare runs dialog
    compare_dialog = ui.dialog().props("full-width")

    # Function to show the side panel with detailed information
    def show_side_panel(run_data: RunData):
        side_panel.clear()
        with side_panel:
            ui.button(">>", on_click=hide_side_panel).style(
                "margin: 20px; margin-bottom: 0px; margin-top: 100px;"
            )
            with ui.card().style(
                "padding: 20px; margin-right: 20px; margin-top: 20px; margin-bottom: 20px; margin-left: 20px"
            ):
                # display run data
                ui.label(f"Step Run ID: {run_data.step_run_id}")
                ui.label(f"Step Name: {run_data.step_name}")
                ui.label("Input Data:").classes("text-lg")
                for key, value in run_data.input_data.items():
                    ui.markdown(f"**{key}**: {value}")
                ui.label("Output Data:").classes("text-lg")
                ui.markdown(f"{run_data.output_data}")
                # rating buttons
                with ui.button_group():
                    datapoint = [
                        datapoint
                        for datapoint in datapoints
                        if datapoint.run_data_id == run_data.step_run_id
                    ][0]
                    ui.button(
                        icon="thumb_up",
                        on_click=lambda: rate_run(datapoint, 1),
                    )
                    ui.button(
                        icon="thumb_down",
                        on_click=lambda: rate_run(datapoint, -1),
                    )

        side_panel.style("transform:translateX(0%);")
        side_panel.update()

    # Function to hide the side panel
    def hide_side_panel():
        side_panel.clear()
        side_panel.style("transform:translateX(100%);")
        side_panel.update()

    # rating function
    def rate_run(datapoint, rating):
        flow.data_store.backend.rate_datapoint(datapoint.id, rating)

    # Function to download the data
    def download_data():
        import json

        logger.info("Downloading data")
        # Get the data
        data = [
            flow.data_store.backend.get_data(datapoint.run_data_id)
            for datapoint in datapoints
        ]
        # create the json file for export
        export_data = []
        for run_data, datapoint in zip(data, datapoints):
            export_data.append(
                {
                    "step_run_id": run_data.step_run_id,
                    "step_name": run_data.step_name,
                    "input_data": run_data.input_data,
                    "output_data": run_data.output_data,
                    "rating": datapoint.rating,
                }
            )
        # download the file
        ui.download(
            src=json.dumps(export_data, indent=4).encode("utf-8"),
            filename=f"{dataset.name}_data.json",
            media_type="application/json",
        )

    # Function to build the UI
    def build_ui():
        runs: List[RunData] = [
            flow.data_store.backend.get_data(datapoint.run_data_id)
            for datapoint in datapoints
        ]

        # Main UI setup
        with ui.card().style("padding: 20px"):
            with ui.row().classes("justify-end"):
                ui.button("Export", on_click=download_data).style("margin-bottom: 20px")
            # Create a table with clickable rows
            columns = [
                {"name": "step_run_id", "label": "step_run_id", "field": "step_run_id"},
                {"name": "name", "label": "name", "field": "name", "sortable": True},
                {
                    "name": "rated",
                    "label": "rated",
                    "field": "rated",
                    "sortable": True,
                },
            ]

            rows = [
                {
                    "step_run_id": run_data.step_run_id,
                    "name": run_data.step_name,
                    "rated": eval_data.rating if eval_data.rating else "Not rated",
                }
                for run_data, eval_data in zip(runs, datapoints)
            ]

            table = ui.table(
                columns=columns,
                rows=rows,
                selection="multiple",
                row_key="step_run_id",
                pagination={
                    "rowsPerPage": 20,
                    "sortBy": "run_time",
                    "page": 1,
                    "descending": True,
                },
            )
            table.add_slot(
                "body-cell-rated",
                """
                <q-td key="rated" :props="props">
                    <q-badge :color="props.value === 'Not rated' ? 'red' : 'green'">
                        {{ props.value }}
                    </q-badge>
                </q-td>
                """,
            )

            def on_row_click(event):
                selected_run_index = event.args[-2]["step_run_id"]
                show_side_panel(
                    run_data=[r for r in runs if r.step_run_id == selected_run_index][
                        -1
                    ]
                )

            table.on("rowClick", on_row_click)

    return build_ui



import inspect
from loguru import logger
from collections import defaultdict, deque
from typing import Dict, Callable, List


# Local imports
from .step import MageStep
from .result import MageResult
from .storage import PromptStore, DataStore, SQLitePromptBackend, SQLiteDataBackend


class PromptMage:
    """A class to represent a PromptMage instance.

    Attributes:
        name (str): The name of the PromptMage instance.
        prompt_store (PromptStore): The prompt store to use for storing prompts.
        data_store (DataStore): The data store to use for storing data.
        steps (Dict): A dictionary of steps in the PromptMage instance.
    """

    def __init__(
        self,
        name: str,
        prompt_store: PromptStore | None = None,
        data_store: DataStore | None = None,
        available_models: List[str] | None = None,
    ):
        self.name: str = name
        self.prompt_store = (
            prompt_store if prompt_store else PromptStore(backend=SQLitePromptBackend())
        )
        self.data_store = (
            data_store if data_store else DataStore(backend=SQLiteDataBackend())
        )
        self.available_models = available_models
        self.steps: Dict[str, MageStep] = {}
        logger.info(f"Initialized PromptMage with name: {name}")

        # store the dependency graph
        self.dependencies = defaultdict(list)
        self.graph = defaultdict(list)
        self.indegree = defaultdict(int)

        # store the pass_through_inputs
        self.pass_through_inputs = {}

        # store execution results and details
        self.execution_results = []

        # store the running status
        self.is_running = False

    def step(
        self,
        name: str,
        prompt_name: str | None = None,
        depends_on: List[str] | str | None = None,
        initial: bool = False,
        one_to_many: bool = False,
        many_to_one: bool = False,
        pass_through_inputs: List[str] | None = None,
    ) -> Callable:
        """Decorator to register a step to the PromptMage instance.

        Args:
            name (str): The name of the step.
            prompt_name (str, optional): The name of the prompt to use for this step. Defaults to None.
            depends_on (str, optional): The name of the step that this step depends on. Defaults to None.
            initial (bool): Whether this step is an initial step. Defaults to False.
            one_to_many (bool, optional): Whether this step is a one-to-many step. Defaults to False.
            many_to_one (bool, optional): Whether this step is a many-to-one step. Defaults to False.
            pass_through_inputs (List[str], optional): The list of inputs to pass through to the step that requires them. Defaults to None.

        One-to-many steps are steps that are expected to return Iterable outputs. Following steps will be executed for each output.
        Many-to-one steps are steps that are expected to receive Iterable inputs. The step will be executed on all inputs together.
        """
        assert not (
            one_to_many and many_to_one
        ), "Cannot be both one-to-many and many-to-one."

        def decorator(func):
            # get the function signature
            func_params = inspect.signature(func).parameters

            # create and store the step
            self.steps[name] = MageStep(
                name=name,
                func=func,
                prompt_store=self.prompt_store,
                data_store=self.data_store,
                prompt_name=prompt_name,
                initial=initial,
                depends_on=depends_on,
                one_to_many=one_to_many,
                many_to_one=many_to_one,
                available_models=(
                    self.available_models if func_params.get("model") else None
                ),
                model=(
                    func_params.get("model").default
                    if func_params.get("model")
                    else None
                ),
            )

            # store the dependencies
            if depends_on:
                self.dependencies[name] = (
                    depends_on if isinstance(depends_on, list) else [depends_on]
                )
            else:
                self.dependencies[name] = []

            return func

        return decorator

    def get_run_function(
        self, start_from: str | None = None, active_prompts: bool | None = None
    ) -> Callable:
        initial_step_name = (
            [step.name for step in self.steps.values() if step.initial][0]
            if not start_from
            else start_from
        )
        first_func_node: MageStep = self.steps[initial_step_name]

        def run_function(**initial_inputs):
            """
            Execute steps starting from the initial step, following the next_step attribute.

            Args:
                initial_inputs (dict): The inputs for the initial step.
            """
            self.is_running = True
            self.execution_results = []

            def execute_graph(
                step_name: str,
                inputs: dict | None = None,
                previous_result_ids: List | None = None,
            ) -> MageResult:
                """
                Helper function to execute a step by its name.
                """
                logger.info(f"Executing step: {step_name}")
                current_node = step_name
                current_data = inputs

                while current_node:
                    logger.warning(f"Current node: {current_node}")
                    # Get the current step
                    step = self.steps[current_node]

                    # Execute the step
                    if isinstance(current_data, list):
                        current_data = combine_dicts(current_data)
                    # check if all required inputs are available else return
                    if not all(
                        input_param in current_data.keys()
                        for input_param in step.signature.parameters
                        if input_param not in ["prompt", "model"]
                    ):
                        logger.warning(
                            f"Step {current_node} requires additional inputs. Skipping."
                        )
                        break
                    response = step.execute(**current_data, active=active_prompts)

                    # Store current and previous result ids
                    if previous_result_ids is None:
                        previous_result_ids = []
                    if isinstance(response, list):
                        for res in response:
                            self.execution_results.append(
                                {
                                    "previous_result_ids": previous_result_ids,
                                    "current_result_id": res.id,
                                    "step": step.name,
                                    "results": res.results,
                                }
                            )
                        previous_result_ids = [res.id for res in response]
                    else:
                        self.execution_results.append(
                            {
                                "previous_result_ids": previous_result_ids,
                                "current_result_id": response.id,
                                "step": step.name,
                                "results": response.results,
                            }
                        )
                        previous_result_ids = [response.id]

                    # Store the execution results
                    if isinstance(response, list):
                        result = response
                        next_node = [r.next_step for r in response]
                    elif isinstance(response, MageResult):
                        next_node = response.next_step
                        result = response
                    else:
                        raise ValueError(
                            f"Step {current_node} returned an invalid response type."
                        )
                    if isinstance(next_node, list):  # Multiple next nodes
                        if isinstance(result, list):
                            logger.warning("Multiple next nodes and multiple results.")
                            # Map the next_node to each result item
                            current_node = next_node
                            current_data = []
                            next_node = None
                            previous_result_ids_list = []
                            for r, n in zip(result, current_node):
                                d, next_node, previous_result_idx = execute_graph(
                                    step_name=n,
                                    inputs=r.results,
                                    previous_result_ids=[r.id],
                                )
                                current_data.append(d)
                                previous_result_ids_list.append(previous_result_idx)
                            previous_result_ids = [
                                id for ids in previous_result_ids_list for id in ids
                            ]
                            current_node = next_node
                        else:
                            logger.warning("Multiple next nodes and single result.")
                            # Passing the single result to each next node
                            current_data = []
                            current_node = next_node
                            next_nodes = []
                            previous_result_ids = []
                            for n in current_node:
                                res, next_node, pid = execute_graph(
                                    step_name=n,
                                    inputs=result.results,
                                    previous_result_ids=[result.id],
                                )
                                previous_result_ids.extend(pid)
                                current_data.append(res)
                                if next_node:
                                    next_nodes.append(next_node)
                            # if all the next node are the same, then we can just use the first one else raise an error
                            if len(set(next_nodes)) == 1:
                                current_node = next_nodes[0]
                            else:
                                logger.error(
                                    f"Step {current_node} returned {next_nodes} as next nodes."
                                )
                                raise ValueError(
                                    "Multiple next nodes and single result. Next nodes are different."
                                )
                    else:
                        if isinstance(result, list):
                            logger.warning("Single next node and multiple results.")
                            current_node = next_node
                            current_data = []
                            next_node = None
                            for r in result:
                                d, next_node, previous_result_ids = execute_graph(
                                    step_name=current_node,
                                    inputs=r.results,
                                    previous_result_ids=[r.id],
                                )
                                current_data.append(d)
                            current_node = next_node
                        else:
                            if next_node and self.steps[next_node].many_to_one:
                                logger.warning(
                                    "Single next node and many-to-one result."
                                )
                                current_data = result.results
                                current_node = next_node
                                break
                            else:
                                logger.warning("Single next node and single result.")
                                current_node = next_node
                                current_data = result.results

                return (
                    current_data,
                    current_node,
                    previous_result_ids if previous_result_ids else None,
                )

            final_result, _, _ = execute_graph(initial_step_name, initial_inputs)
            self.is_running = False
            return final_result

        # Set the signature of the returned function to match the first function in the graph
        run_function.__signature__ = first_func_node.signature
        return run_function

    def __repr__(self) -> str:
        return f"PromptMage(name={self.name}, steps={list(self.steps.keys())})"

    def get_run_data(self) -> List:
        if self.data_store:
            runs = self.data_store.get_all_data()
            # filter by step_name
            return [run for run in runs if run.step_name in self.steps]
        return []


def combine_dicts(list_of_dicts):
    # Initialize a defaultdict where each key will hold a list of values
    combined_dict = defaultdict(list)

    # Iterate through each dictionary in the list
    for d in list_of_dicts:
        for key, value in d.items():
            combined_dict[key].append(value)

    # Convert lists with a single value back to that value
    final_dict = {
        key: (value[0] if len(value) == 1 else value)
        for key, value in combined_dict.items()
    }

    return final_dict



from nicegui import ui
from loguru import logger
from collections import defaultdict
from typing import List

from promptmage import PromptMage
from promptmage.prompt import Prompt
from .styles import label_with_icon


def create_prompts_view(mage: PromptMage):
    side_panel = (
        ui.element("div")
        .style(
            "position: fixed; top: 0; right: 0; width: 50%; height: 100%; transform: translateX(100%); transition: transform 0.3s ease; z-index: 1000; overflow-y: auto;"
        )
        .classes("bg-gray-100 dark:bg-slate-800")
    )

    # prompt editing dialog
    dialog = ui.dialog().props("full-width")

    # Function to show the side panel with detailed information
    def show_side_panel(prompts: List[Prompt]):
        # get prompt with highest version
        side_panel.clear()
        with side_panel:
            ui.button(">>", on_click=hide_side_panel).style(
                "margin: 20px; margin-bottom: 0px; margin-top: 100px;"
            )
            for prompt in sorted(prompts, key=lambda p: p.version, reverse=True):
                bg_color = ""
                # highlight active prompt in green
                if prompt.active:
                    bg_color = "bg-green-200 dark:bg-green-800"
                with ui.card().style(
                    "padding: 20px; margin-right: 20px; margin-top: 10px; margin-bottom: 10px; margin-left: 20px;"
                ).classes(bg_color):
                    # display run data
                    with ui.grid(columns=2).classes("gap-0"):
                        label_with_icon("Prompt ID:", icon="o_info")
                        ui.label(f"{prompt.id}")

                        label_with_icon("Name:", icon="o_badge")
                        ui.label(f"{prompt.name}")

                        label_with_icon("Version:", icon="o_tag")
                        ui.label(f"{prompt.version}")

                        label_with_icon("Active:", icon="o_check")
                        ui.label(f"{prompt.active}")

                        label_with_icon("System prompt:", icon="o_code")
                        ui.label(f"{prompt.system}")

                        label_with_icon("User prompt:", icon="o_psychology")
                        ui.label(f"{prompt.user}")

                    with ui.row():
                        ui.button(
                            "Activate Prompt",
                            on_click=lambda prompt_id=prompt.id: activate_prompt(
                                prompt_id
                            ),
                        )
                        ui.button(
                            "Edit Prompt",
                            on_click=lambda prompt_id=prompt.id: edit_prompt(prompt_id),
                        )
                        ui.button(
                            "Delete Prompt",
                            on_click=lambda prompt_id=prompt.id: delete_prompt(
                                prompt_id
                            ),
                        )
        side_panel.style("transform:translateX(0%);")
        side_panel.update()

    def delete_prompt(prompt_id):
        logger.info(f"Deleting prompt with ID: {prompt_id}.")
        mage.prompt_store.delete_prompt(prompt_id)
        ui.notify(f"Prompt {prompt_id} deleted.")
        side_panel.update()

    def edit_prompt(prompt_id):
        logger.info(f"Editing prompt with ID: {prompt_id}.")
        prompt = mage.prompt_store.get_prompt_by_id(prompt_id)
        dialog.clear()
        with dialog, ui.card():
            ui.label("Edit prompt").classes("text-2xl")
            with ui.row():
                ui.label(f"Name: {prompt.name}")
                ui.label(f"Version: {prompt.version}")
                ui.label(f"Prompt ID: {prompt_id}")
            with ui.row():
                system_prompt = ui.textarea(
                    value=prompt.system, label="System prompt"
                ).style("width: 500px; height: 200px;")
                user_prompt = ui.textarea(value=prompt.user, label="User prompt").style(
                    "width: 500px; height: 200px;"
                )
            with ui.row():
                ui.button(
                    "Save",
                    on_click=lambda: save_prompt(
                        prompt_id, system_prompt.value, user_prompt.value
                    ),
                )
                ui.button("Cancel", on_click=dialog.close)

        dialog.open()

    def activate_prompt(prompt_id):
        logger.info(f"Activating prompt with ID: {prompt_id}.")
        # activate the selected prompt
        prompt = mage.prompt_store.get_prompt_by_id(prompt_id)
        prompt.active = True
        mage.prompt_store.update_prompt(prompt)
        # deactivate all other prompts with the same name
        for p in mage.prompt_store.get_prompts():
            if p.name == prompt.name and p.id != prompt_id:
                p.active = False
                mage.prompt_store.update_prompt(p)
        ui.notify(f"Prompt {prompt_id} activated.")
        side_panel.update()

    def save_prompt(prompt_id: str, system: str, user: str):
        prompt = mage.prompt_store.get_prompt_by_id(prompt_id)
        prompt.system = system
        prompt.user = user
        mage.prompt_store.update_prompt(prompt)
        dialog.close()

    # Function to hide the side panel
    def hide_side_panel():
        side_panel.clear()
        side_panel.style("transform:translateX(100%);")
        side_panel.update()

    def build_ui():
        all_prompts = mage.prompt_store.get_prompts()
        # group them by name
        prompts = defaultdict(list)
        for prompt in all_prompts:
            if prompt.name in [step.prompt_name for step in mage.steps.values()]:
                prompts[prompt.name].append(prompt)

        # Main UI setup
        with ui.card().style("padding: 20px"):
            # Create a table with clickable rows
            columns = [
                {"name": "name", "label": "name", "field": "name", "sortable": True},
                {
                    "name": "versions",
                    "label": "Number of versions",
                    "field": "versions",
                    "sortable": True,
                },
            ]

            rows = [
                {
                    "name": name,
                    "versions": len(prompts_for_name),
                }
                for name, prompts_for_name in prompts.items()
            ]

            table = ui.table(columns=columns, rows=rows)

            def on_row_click(event):
                selected_run_index = event.args[-2]["name"]
                show_side_panel(prompts=prompts[selected_run_index])

            table.on("rowClick", on_row_click)

    return build_ui



"""This module contains the StorageBackend class, which is an abstract class that defines the interface for storage backends."""

from abc import ABC, abstractmethod
from pathlib import Path


class StorageBackend(ABC):
    """An abstract class that defines the interface for storage backends."""



from nicegui import Tailwind, ui


textbox_style = "w-full max-w-3xl"  #  border-2 border-gray-300 rounded-lg ring focus:outline-none focus:ring-2 focus:ring-blue-500 focus:border-transparent"


def label_with_icon(label: str, icon: str):
    row = ui.row().classes("gap-2 items-center")
    with row:
        ui.icon(icon)
        ui.label(label)
    return row



"""This ui element represent the input, prompt and output of a callable step in the PromptMage."""

import inspect
from contextlib import contextmanager
from nicegui import ui, run

from promptmage import PromptMage
from .styles import textbox_style


@contextmanager
def create_main_runner(mage: PromptMage, execution_graph):
    input_fields = {}
    result_field = None
    flow_func = mage.get_run_function(start_from=None)

    async def run_function():
        inputs = {name: field.value for name, field in input_fields.items()}
        mage.is_running = True
        execution_graph.refresh()
        result = await run.io_bound(flow_func, **inputs)
        newline = "\n\n"
        if isinstance(result, list):
            result_field.set_content(
                f"{[newline.join(res.values()) for res in result]}"
            )
        else:
            result_field.set_content(f"{newline.join(result.values())}")
        result_field.update()
        execution_graph.refresh()

    def build_ui():
        nonlocal result_field
        with ui.column().classes("w-full"):
            # elements before the steps runner
            ui.label("Inputs:").classes("font-bold text-lg")
            for param in inspect.signature(flow_func).parameters.values():
                if param.name not in ["prompt", "model"]:
                    with ui.row().classes("w-full"):
                        input_fields[param.name] = (
                            ui.textarea(label=f"{param.name}")
                            .classes(textbox_style)
                            .props("outlined")
                        )

            with ui.row().classes("w-full justify-end"):
                ui.button("Run", on_click=run_function)
            ui.separator()
            # steps runner
            ui.label("Steps:").classes("font-bold text-lg")
            yield
            # elements after the steps runner
            ui.separator()
            with ui.row().classes("w-full justify-between"):
                ui.label("Result:").classes("font-bold text-lg")
                ui.button(
                    "Copy to clipboard",
                    on_click=lambda: ui.clipboard.write(result_field.content),
                )  # .props("fab")
            result_field = (
                ui.markdown("")
                .style("height: 200px; overflow-y: auto;")
                .classes("color-black dark:color-white")
            )

    return build_ui()



"""This module contains the SQLiteBackend class, which is a subclass of the StorageBackend class. It is used to store the data in a SQLite database."""

import json
import uuid
from loguru import logger
from typing import List, Dict
from sqlalchemy import (
    create_engine,
    Column,
    String,
    Integer,
    Text,
    select,
    delete,
    DateTime,
    UUID,
    ForeignKey,
    Boolean,
    text,
    and_,
    Float,
)
from sqlalchemy.sql import func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.exc import SQLAlchemyError

from promptmage.prompt import Prompt
from promptmage.exceptions import PromptNotFoundException
from promptmage.run_data import RunData
from promptmage.storage.storage_backend import StorageBackend

Base = declarative_base()


def generate_uuid():
    return str(uuid.uuid4())


class PromptModel(Base):
    __tablename__ = "prompts"
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    system = Column(Text, nullable=False)
    user = Column(Text, nullable=False)
    version = Column(Integer, nullable=False)
    template_vars = Column(Text, nullable=False)
    active = Column(Boolean, nullable=False, default=False)

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "version": self.version,
            "active": self.active,
            "system": self.system,
            "user": self.user,
            "template_vars": self.template_vars.split(","),
        }

    @classmethod
    def from_dict(cls, data: Dict) -> "PromptModel":
        return cls(
            id=data["id"],
            name=data["name"],
            system=data["system"],
            user=data["user"],
            version=data["version"],
            active=data["active"],
            template_vars=",".join(data["template_vars"]),
        )

    def __repr__(self):
        return (
            f"PromptModel(id={self.id}, "
            f"name={self.name}, "
            f"active={self.active}, "
            f"system={self.system}, "
            f"user={self.user}, "
            f"version={self.version}, "
            f"template_vars={self.template_vars})"
        )


class SQLitePromptBackend(StorageBackend):
    """A class that stores the prompts in a SQLite database.

    Attributes:
        db_path (str): The path to the SQLite database. Defaults to ".promptmage/promptmage.db".
    """

    def __init__(self, db_path: str | None = None):
        self.db_path = db_path if db_path else ".promptmage/promptmage.db"
        self.engine = create_engine(f"sqlite:///{self.db_path}")
        Base.metadata.create_all(self.engine)

        # Define the SQL command to update the existing rows
        # update_command = text("UPDATE prompts SET active = false WHERE active IS NULL")
        # Execute the update command
        # with self.engine.connect() as conn:
        #     conn.execute(update_command)

        self.Session = sessionmaker(bind=self.engine)

    def store_prompt(self, prompt: Prompt):
        session = self.Session()
        try:
            session.add(PromptModel.from_dict(prompt.to_dict()))
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error storing prompt: {e}")
        finally:
            session.close()

    def update_prompt(self, prompt: Prompt):
        """Update an existing prompt in the database by id.

        Args:
            prompt (Prompt): The prompt to update.
        """
        session = self.Session()
        try:
            existing_prompt = session.execute(
                select(PromptModel).where(PromptModel.id == prompt.id)
            ).scalar_one_or_none()

            if not existing_prompt:
                raise PromptNotFoundException(
                    f"Prompt with name {prompt.name} not found."
                )

            existing_prompt.system = prompt.system
            existing_prompt.user = prompt.user
            existing_prompt.active = prompt.active

            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error updating prompt: {e}")
        finally:
            session.close()

    def get_prompt(
        self, prompt_name: str, version: int | None = None, active: bool | None = None
    ) -> Prompt:
        """Get a prompt by name from the database.

        Args:
            prompt_name (str): The name of the prompt to retrieve.
            version (int): The version of the prompt to retrieve.
            active (bool): Whether to retrieve only the active prompt.
        """
        session = self.Session()
        try:
            # build the where clause based on name version and active
            where_clause = [PromptModel.name == prompt_name]
            if version is not None:
                where_clause.append(PromptModel.version == version)
            if active is not None:
                where_clause.append(PromptModel.active == active)
            combined_where_clause = and_(*where_clause)
            # run the query to get the prompt
            rows = (
                session.execute(select(PromptModel).where(combined_where_clause))
                .scalars()
                .all()
            )
            if not rows:
                raise PromptNotFoundException(
                    f"Prompt with name {prompt_name} not found."
                )
            latest_prompt = max(rows, key=lambda p: p.version)
            return Prompt(**latest_prompt.to_dict())
        finally:
            session.close()

    def get_prompt_by_id(self, prompt_id: str) -> Prompt:
        session = self.Session()
        try:
            prompt = session.execute(
                select(PromptModel).where(PromptModel.id == prompt_id)
            ).scalar_one_or_none()
            if prompt is None:
                raise PromptNotFoundException(f"Prompt with ID {prompt_id} not found.")
            return Prompt(**prompt.to_dict())
        finally:
            session.close()

    def get_prompts(self) -> List[Prompt]:
        session = self.Session()
        try:
            prompts = session.execute(select(PromptModel)).scalars().all()
            return [Prompt(**prompt.to_dict()) for prompt in prompts]
        finally:
            session.close()

    def delete_prompt(self, prompt_id: str):
        session = self.Session()
        try:
            result = session.execute(
                delete(PromptModel).where(PromptModel.id == prompt_id)
            )
            if result.rowcount == 0:
                raise PromptNotFoundException(f"Prompt with ID {prompt_id} not found.")
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error deleting prompt: {e}")
        finally:
            session.close()


class RunDataModel(Base):
    __tablename__ = "data"
    step_run_id = Column(String, primary_key=True)
    run_time = Column(String, nullable=False)
    execution_time = Column(Float, nullable=True)
    model = Column(String, nullable=True)
    step_name = Column(String, nullable=False)
    run_id = Column(String)
    status = Column(String, nullable=False)
    prompt = Column(Text)
    input_data = Column(Text)
    output_data = Column(Text)

    def to_dict(self) -> Dict:
        return {
            "step_run_id": self.step_run_id,
            "run_time": self.run_time,
            "step_name": self.step_name,
            "run_id": self.run_id,
            "status": self.status,
            "execution_time": self.execution_time,
            "model": self.model,
            "prompt": Prompt(**json.loads(self.prompt)) if self.prompt else None,
            "input_data": json.loads(self.input_data),
            "output_data": json.loads(self.output_data),
        }

    @classmethod
    def from_dict(cls, data: Dict) -> "RunDataModel":
        return cls(
            step_run_id=data["step_run_id"],
            run_time=data["run_time"],
            step_name=data["step_name"],
            run_id=data["run_id"],
            status=data["status"],
            execution_time=data["execution_time"],
            model=data["model"],
            prompt=json.dumps(data["prompt"]) if data["prompt"] else None,
            input_data=json.dumps(data["input_data"]),
            output_data=json.dumps(data["output_data"]),
        )

    def __repr__(self):
        return (
            f"RunDataModel(step_run_id={self.step_run_id}, "
            f"run_time={self.run_time}, "
            f"step_name={self.step_name}, "
            f"run_id={self.run_id}, "
            f"status={self.status}, "
            f"prompt={self.prompt}, "
            f"execution_time={self.execution_time}, "
            f"model={self.model}, "
            f"input_data={self.input_data}, "
            f"output_data={self.output_data})"
        )


class EvaluationDatasetModel(Base):
    __tablename__ = "evaluation_datasets"
    id = Column("id", String, primary_key=True, default=generate_uuid)
    name = Column(String, nullable=False)
    created = Column(DateTime(timezone=True), server_default=func.now())
    updated = Column(DateTime(timezone=True), onupdate=func.now())
    datapoints = relationship(
        "EvaluationDatapointModel",
        back_populates="dataset",
        cascade="all, delete-orphan",
    )

    @classmethod
    def from_dict(cls, data: Dict) -> "EvaluationDatasetModel":
        return cls(
            id=data["id"],
            name=data["name"],
        )

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
        }

    def __repr__(self):
        return f"EvaluationDatasetModel(id={self.id}, name={self.name})"


class EvaluationDatapointModel(Base):
    __tablename__ = "evaluation_datapoints"
    id = Column("id", String, primary_key=True, default=generate_uuid)
    dataset_id = Column(String, ForeignKey("evaluation_datasets.id"))
    dataset = relationship("EvaluationDatasetModel", back_populates="datapoints")
    run_data_id = Column(String, ForeignKey("data.step_run_id"))
    rating = Column(Integer, nullable=True, default=None)

    @classmethod
    def from_dict(cls, data: Dict) -> "EvaluationDatapointModel":
        return cls(
            id=data["id"],
            dataset_id=data["dataset_id"],
            run_data_id=data["run_data_id"],
            rating=data.get("rating"),
        )

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "dataset_id": self.dataset_id,
            "run_data_id": self.run_data_id,
            "rating": self.rating,
        }

    def __repr__(self):
        return f"EvaluationDatapointModel(id={self.id}, dataset_id={self.dataset_id}, run_data_id={self.run_data_id}, rating={self.rating})"


class SQLiteDataBackend(StorageBackend):
    """A class that stores the data in a SQLite database.

    Attributes:
        db_path (str): The path to the SQLite database. Defaults to ".promptmage/promptmage.db".
    """

    def __init__(self, db_path: str | None = None):
        self.db_path = db_path if db_path else ".promptmage/promptmage.db"
        self.engine = create_engine(f"sqlite:///{self.db_path}")
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)

    def store_data(self, run_data: RunData):
        session = self.Session()
        try:
            session.add(RunDataModel.from_dict(run_data.to_dict()))
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error storing run data: {e}")
        finally:
            session.close()

    def get_data(self, step_run_id: str) -> RunData:
        session = self.Session()
        try:
            run_data = session.execute(
                select(RunDataModel).where(RunDataModel.step_run_id == step_run_id)
            ).scalar_one_or_none()
            if run_data is None:
                return None
            return RunData(**run_data.to_dict())
        finally:
            session.close()

    def get_all_data(self) -> List[RunData]:
        session = self.Session()
        try:
            run_data_list = session.execute(select(RunDataModel)).scalars().all()
            return [RunData(**d.to_dict()) for d in run_data_list]
        finally:
            session.close()

    def create_dataset(self, name: str):
        session = self.Session()
        try:
            dataset = EvaluationDatasetModel(name=name)
            session.add(dataset)
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error creating dataset: {e}")
        finally:
            session.close()

    def delete_dataset(self, dataset_id: str):
        session = self.Session()
        try:
            result = session.execute(
                delete(EvaluationDatasetModel).where(
                    EvaluationDatasetModel.id == dataset_id
                )
            )
            if result.rowcount == 0:
                raise ValueError(f"Dataset with ID {dataset_id} not found.")
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error deleting dataset: {e}")
        finally:
            session.close()

    def add_datapoint_to_dataset(self, datapoint_id, dataset_id):
        session = self.Session()
        try:
            datapoint = EvaluationDatapointModel(
                run_data_id=datapoint_id, dataset_id=dataset_id
            )
            session.add(datapoint)
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error adding datapoint to dataset: {e}")
        finally:
            session.close()

    def get_datasets(self) -> List[EvaluationDatasetModel]:
        session = self.Session()
        try:
            datasets = session.execute(select(EvaluationDatasetModel)).scalars().all()
            logger.info(f"Datasets: {datasets}")
            return [d for d in datasets]
        finally:
            session.close()

    def get_dataset(self, dataset_id: str) -> EvaluationDatasetModel:
        session = self.Session()
        try:
            dataset = session.execute(
                select(EvaluationDatasetModel).where(
                    EvaluationDatasetModel.id == dataset_id
                )
            ).scalar_one_or_none()
            if dataset is None:
                raise ValueError(f"Dataset with ID {dataset_id} not found.")
            return dataset
        finally:
            session.close()

    def get_datapoints(self, dataset_id: str) -> List[EvaluationDatapointModel]:
        session = self.Session()
        try:
            datapoints = (
                session.execute(
                    select(EvaluationDatapointModel).where(
                        EvaluationDatapointModel.dataset_id == dataset_id
                    )
                )
                .scalars()
                .all()
            )
            return [d for d in datapoints]
        finally:
            session.close()

    def get_datapoint(self, datapoint_id: str) -> EvaluationDatapointModel:
        session = self.Session()
        try:
            datapoint = session.execute(
                select(EvaluationDatapointModel).where(
                    EvaluationDatapointModel.id == datapoint_id
                )
            ).scalar_one_or_none()
            if datapoint is None:
                raise ValueError(f"Datapoint with ID {datapoint_id} not found.")
            return datapoint
        finally:
            session.close()

    def rate_datapoint(self, datapoint_id: str, rating: int):
        session = self.Session()
        try:
            datapoint = session.execute(
                select(EvaluationDatapointModel).where(
                    EvaluationDatapointModel.id == datapoint_id
                )
            ).scalar_one_or_none()
            if datapoint is None:
                raise ValueError(f"Datapoint with ID {datapoint_id} not found.")
            datapoint.rating = rating
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error rating datapoint: {e}")
        finally:
            session.close()

    def remove_datapoint_from_dataset(self, datapoint_id: str, dataset_id: str):
        session = self.Session()
        try:
            result = session.execute(
                delete(EvaluationDatapointModel).where(
                    EvaluationDatapointModel.id == datapoint_id
                )
            )
            if result.rowcount == 0:
                raise ValueError(f"Datapoint with ID {datapoint_id} not found.")
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            logger.error(f"Error removing datapoint from dataset: {e}")
        finally:
            session.close()



"""This module contains the command line interface for the PromptMage package."""

import json
import click
import uvicorn
from pathlib import Path
from loguru import logger

from promptmage import __version__, title
from promptmage.utils import get_flows
from promptmage.api import PromptMageAPI
from promptmage.frontend import PromptMageFrontend
from promptmage.storage import SQLiteDataBackend, SQLitePromptBackend


@click.group()
def promptmage():
    """Promptmage CLI"""
    pass


@click.command()
def version():
    """Print the version of the PromptMage package."""
    click.echo(f"PromptMage version: {__version__}")


@click.command()
@click.argument(
    "file_path",
    type=click.Path(
        exists=True,
    ),
)
@click.option("--host", default="0.0.0.0", help="The host IP to run the server on.")
@click.option("--port", default=8000, type=int, help="The port to run the server on.")
@click.option(
    "--browser",
    is_flag=True,
    help="Open the browser after starting the server.",
    default=False,
)
def run(file_path: str, host: str, port: int, browser: bool):
    """Serve the application containing a PromptMage instance from the given file.

    Args:
        file_path (str): The path to the file containing the PromptMage instance.
        host (str): The host IP to run the FastAPI server on.
        port (int): The port to run the FastAPI server on.
        browser (bool): Whether to open the browser after starting the server.
    """
    logger.info(f"\nWelcome to\n{title}")
    logger.info(f"Running PromptMage version {__version__} from {file_path}")
    # create the .promptmage directory to store all the data
    dirPath = Path(".promptmage")
    dirPath.mkdir(mode=0o777, parents=False, exist_ok=True)

    # get the available flows from the source file
    available_flows = get_flows(file_path)

    if not available_flows:
        raise ValueError("No PromptMage instance found in the module.")

    # create the FastAPI app
    app = PromptMageAPI(flows=available_flows).get_app()

    # create the frontend app
    frontend = PromptMageFrontend(flows=available_flows)
    frontend.init_from_api(app)

    # Run the applications
    if browser:
        import webbrowser

        url = f"http://localhost:{port}"
        webbrowser.open_new_tab(url)
    uvicorn.run(app, host=host, port=port, log_level="info")


@click.command()
@click.option("--runs", "runs", default=False, help="Export runs.", flag_value=True)
@click.option(
    "--prompts", "prompts", default=False, help="Export prompts.", flag_value=True
)
@click.option(
    "--filename",
    default="promptmage",
    help="The name of the file to export the data to.",
)
def export(runs: bool = False, prompts: bool = False, filename: str = "promptmage"):
    """Export the run data and prompts from the PromptMage instance to json.

    Args:
        runs (bool): Whether to export the run data.
        prompts (bool): Whether to export the prompts.
        filename (str): The name of the file to export the data to.
    """
    if runs:
        click.echo("Exporting runs...")
        data_store = SQLiteDataBackend()
        run_data = data_store.get_all_data()

        with open(f"{filename}_runs.json", "w") as f:
            json.dump([run for run in run_data.values()], f)

    if prompts:
        click.echo("Exporting prompts...")
        prompt_store = SQLitePromptBackend()
        prompts = prompt_store.get_prompts()

        with open(f"{filename}_prompts.json", "w") as f:
            json.dump([prompt.to_dict() for prompt in prompts], f)

    if not runs and not prompts:
        click.echo("No data to export.")
    else:
        click.echo("Export complete.")


promptmage.add_command(version)
promptmage.add_command(run)
promptmage.add_command(export)


if __name__ == "__main__":
    promptmage()



import sys
from pathlib import Path
from typing import List
from importlib import import_module

from promptmage import PromptMage


def get_flows(file_path: str) -> List[PromptMage]:
    sys.path.append(
        str(Path(file_path).parent.absolute())
    )  # Add the directory of the file to PYTHONPATH
    module_name = Path(file_path).stem
    module = import_module(module_name)

    # Find an instances of PromptMage in the module
    flows = []
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if isinstance(attr, PromptMage):
            flows.append(attr)
    return flows



import uuid


class MageResult:

    def __init__(
        self,
        next_step: str | None = None,
        error: str | None = None,
        **kwargs,
    ):
        self.id = str(uuid.uuid4())
        self.next_step = next_step
        self.results: dict = kwargs
        self.error = error

    def __repr__(self):
        return f"<MageResult id={self.id} next_step={self.next_step} results={self.results} error={self.error}>"
